from collections import OrderedDict
from typing import List, Dict, Callable, Union, Any, Type, Tuple
from .models import JobModel
from .parsers import DefaultParser
from .exporters import DefaultExporter


class Engine:

    job_model_class: Type[JobModel] = JobModel
    parser_class: Type[DefaultParser] = DefaultParser
    exporter_class: Type[DefaultExporter] = DefaultExporter

    parsed_rule: List[JobModel] = []

    session: Dict[str, Any] = {}

    def __init__(
        self,
        context: Dict[str, Any] = None,
        job_model: Type[JobModel] = None,
        parser_class: Type[DefaultParser] = None,
        exporter_class: Type[DefaultExporter] = None,
    ):
        """
        - Sessions can be initialized with a context provided by the user
        - Job Model and Parser can be changed
        """
        if context:
            self.session = context

        if job_model:
            self.job_model_class = job_model

        if parser_class:
            self.parser_class = parser_class

        if exporter_class:
            self.exporter_class = exporter_class
        self.callables_collected: "OrderedDict[str, Dict[str, Any]]" = OrderedDict()

    def _add_callable(self, function: Callable, verbose_name: str, job_type: str = "job"):
        self.callables_collected[function.__name__] = {
            "function": function,
            "verbose_name": verbose_name,
            "type": job_type
        }

    def job(self, *args, **kwargs):
        """
        Decorator so jobs can be called as follows:
        @obj.job
        def my_function():
            return None
        """

        def decorator(function: Callable):
            verbose_name: str = kwargs.get("verbose_name", None)
            if args:
                verbose_name = args[0]
            self._add_callable(function, verbose_name)
            return function

        return decorator

    def apply_job_call(
        self, job: JobModel, session: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Any]:
        target_func: Callable = self.callables_collected.get(job.name).get(
            "function"
        )
        if job.args:
            result = target_func(session=session, **job.args)
        else:
            result = target_func(session=session)
        # append result of function called into session
        results = session.get("results", None)
        if not results:
            session["results"] = []
        session["results"].append({"job": job.name, "return": result})
        self.session = session
        return (session, result)

    def parse(self, unparsed_rule: Union[str, Dict[str, Any]]):
        """
            Parses rules
        """
        parser: DefaultParser = self.parser_class()
        parsed_rule: List[JobModel] = parser.parse(unparsed_rule)
        self.parsed_rule = parsed_rule
        return parsed_rule

    def run(
        self, rule: Union[str, Dict[str, Any]], session: Dict[str, Any] = None
    ):
        """
        Executes each job passing the current session to them
        """

        if not session:
            session = self.session

        for job in self.parse(rule):
            session, result = self.apply_job_call(job, session)
            if not result:
                break

    def export_metadata(self, fmt: str = "dict"):
        exporter = self.exporter_class()
        return exporter.export_jobs(self.callables_collected, fmt=fmt)



from typing import List, Dict, Any

from pydantic import BaseModel


# Generic Models
class JobModel(BaseModel):
    job_type: str = "job"
    name: str
    args: Dict[str, Any] = None


# Rule Engine Models
class ConditionModel(JobModel):
    job_type: str = "condition"


class ActionModel(JobModel):
    job_type: str = "action"


class RuleModel(BaseModel):
    conditions: List[ConditionModel]
    actions: List[ActionModel]



import json
from json.decoder import JSONDecodeError
from typing import List, Type, Union, Dict, Any
from sauron.models import JobModel
from ruamel.yaml import YAML


class DefaultParser:
    single_model: Type[JobModel] = JobModel

    def __init__(self):
        self.yaml = YAML(typ="safe")

    def _parse_single_job(self, job_dict) -> JobModel:
        """
        Method that know how to parse a single job dictionary
        """
        return self.single_model(**job_dict)

    def _parse_jobs_from_list(self, jobs_input) -> List[JobModel]:
        """
        Method that know how to parse a list for jobs
        """
        parsed_jobs: List = []
        for raw_job in jobs_input:
            current_job: JobModel = self._parse_single_job(raw_job)
            parsed_jobs.append(current_job)
        return parsed_jobs

    def _parse_jobs_from_string(self, jobs_input) -> List[JobModel]:
        """
        Method that know how to parse a list for jobs described by a
        json-string with the list of jobs
        """
        try:
            jobs: list = self.yaml.load(jobs_input)
        except JSONDecodeError:
            raise ValueError("jobs param is not a valid json string")
        else:
            return self._parse_jobs_from_list(jobs)

    def parse(self, jobs_input) -> List[JobModel]:
        """
        Main method called to parse any jobs
        """
        jobs_list_data: List[JobModel] = []
        if isinstance(jobs_input, str):
            jobs_list_data = self._parse_jobs_from_string(jobs_input)
        elif isinstance(jobs_input, list):
            # jobs_input is a python list
            jobs_list_data = self._parse_jobs_from_list(jobs_input)
        else:
            raise ValueError("jobs param must be a list or json-string")
        return jobs_list_data


class RuleEngineParser(DefaultParser):
    single_model: Type[JobModel] = JobModel

    def __init__(self):
        self.yaml = YAML(typ="safe")

    def _parse_jobs_from_string(self, jobs_input: str) -> List[JobModel]:
        """
        Method that know how to parse a list for jobs described by a
        json-string with the list of jobs
        """
        try:
            decoded_jobs: dict = self.yaml.load(jobs_input)
            jobs: list = decoded_jobs["conditions"] + decoded_jobs["actions"]
        except JSONDecodeError:
            raise ValueError("jobs param is not a valid json string")
        else:
            return self._parse_jobs_from_list(jobs)

    def parse(self, jobs_input: Union[List, str]) -> List[JobModel]:
        """
        Main method called to parse any jobs
        """
        jobs_list_data: List[JobModel] = []
        if isinstance(jobs_input, str):
            jobs_list_data = self._parse_jobs_from_string(jobs_input)
        elif isinstance(jobs_input, list):
            # jobs_input is a python list
            jobs_list_data = self._parse_jobs_from_list(jobs_input)
        else:
            raise ValueError("jobs param must be a list or json-string")
        return jobs_list_data



from collections import OrderedDict
from typing import List, Dict, Callable, Union, Any, Type
from .models import JobModel
from enum import Enum
import inspect
import json as json_lib
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO


class MyYAML(YAML):
    """
        ruamel.yaml lib was chosen as it supports yaml1.2, this gives us json parsing
        together with yaml parsing
        Allows to dump yaml to a string. This can/should be reviewed as this is less
        performatic than writing to a file or stdout according to docs
        https://yaml.readthedocs.io/en/latest/example.html#output-of-dump-as-a-string
    """

    def dump(self, data, stream=None, **kw):
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


class DefaultExporter:
    def __init__(self):
        self.yaml: Type[YAML] = MyYAML(typ="safe")

    def get_job_types(self):
        # TODO: implement this method to get the job types available
        ...

    @staticmethod
    def get_param_info(param):
        """
        Get Type and Defaults of the parameter. If the parameter is an Enum,
        also gets the choices available
        """
        annotation = param.annotation
        name = annotation.__name__
        choices = None
        if Enum in annotation.__mro__:
            choices = [choice for choice in annotation.__members__]
        defaults = param.default
        if defaults is param.empty:
            defaults = None
        return name, choices, defaults

    @classmethod
    def _get_function_metadata(
        cls, input_function: Callable
    ) -> Dict[str, Any]:
        """
            Metadata about arguments documentation and the function itself
        """
        signature: inspect.Signature = inspect.signature(input_function)
        arguments_metadata: Dict[str, Dict[str, Any]] = {}
        for key, parameter in signature.parameters.items():
            arg_type, arg_choices, arg_defaults = cls.get_param_info(parameter)
            arguments_metadata[key] = {
                "default": arg_defaults,
                "type": arg_type,
                "choices": arg_choices,
            }

        metadata = {
            "args": arguments_metadata,
            "doc": inspect.getdoc(input_function),
            "name": input_function.__name__,
        }

        return metadata

    def get_metadata(self, input_function: Dict[str, Any]) -> Dict[str, Any]:
        metadata = self._get_function_metadata(input_function["function"])
        verbose_name = input_function.get("verbose_name", None)
        if verbose_name:
            metadata["name"] = verbose_name
        metadata["type"] = input_function.get("type", "job")

        return metadata

    def export_job(self, jobs: Dict[str, Any]) -> Dict[str, Any]:
        result = {}
        for name, item in jobs.items():
            result[name] = self.get_metadata(item)
        return result

    def export_jobs(
        self, data: Dict[str, Callable], fmt: str = "dict"
    ) -> Union[str, Dict[str, Any]]:
        jobs = self.export_job(data)
        if fmt == "json":
            return json_lib.dumps(jobs)
        elif fmt == "yaml":
            return self.yaml.dump(jobs)
        else:
            return jobs


class RuleEngineExporter(DefaultExporter):
    def __init__(self):
        self.yaml: Type[YAML] = MyYAML(typ="safe")

    def export_job(self, jobs: Dict[str, Any]) -> Dict[str, Any]:
        result = super().export_job(jobs)
        # Separar jobs por tipo
        result_splited_by_type = {}
        for key, value in result.items():
            current_type = result_splited_by_type.get(value["type"], {})
            current_type[key] = value
            result_splited_by_type[value["type"]] = current_type
        print(result_splited_by_type)
        return result_splited_by_type

    def export_jobs(
        self, data: Dict[str, Callable], fmt: str = "dict"
    ) -> Union[str, Dict[str, Any]]:
        jobs = self.export_job(data)
        if fmt == "json":
            return json_lib.dumps(jobs)
        elif fmt == "yaml":
            return self.yaml.dump(jobs)
        else:
            return jobs



from .engine import Engine
from typing import Callable, Type
from .parsers import DefaultParser, RuleEngineParser
from .exporters import DefaultExporter, RuleEngineExporter


class RuleEngine(Engine):

    parser_class: Type[DefaultParser] = RuleEngineParser
    exporter_class: Type[DefaultExporter] = RuleEngineExporter

    def condition(self, *args, **kwargs):
        """
        Decorator so jobs can be called as follows:
        @obj.condition()
        def my_function():
            return None
        """

        def decorator(function: Callable):
            verbose_name: str = kwargs.get("verbose_name", None)
            if args:
                verbose_name = args[0]
            self._add_callable(function, verbose_name, job_type="condition")
            return function

        return decorator

    def action(self, *args, **kwargs):
        """
        Decorator so jobs can be called as follows:
        @obj.action()
        def my_function():
            return None
        """

        def decorator(function: Callable):
            verbose_name: str = kwargs.get("verbose_name", None)
            if args:
                verbose_name = args[0]
            self._add_callable(function, verbose_name, job_type="action")
            return function

        return decorator



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['sauron', 'sauron_rule_engine']

package_data = \
{'': ['*']}

install_requires = \
['mypy>=0.711.0,<0.712.0',
 'pydantic>=0.29.0,<0.30.0',
 'ruamel.yaml>=0.16.5,<0.17.0']

setup_kwargs = {
    'name': 'sauron-rule-engine',
    'version': '1.0.2',
    'description': 'A simple rule engine implemented in python',
    'long_description': '<h1 align="center">Sauron Rule engine - One engine to rule them all </h1>\n<p>\n  <img src="https://img.shields.io/badge/version-0.1-blue.svg?cacheSeconds=2592000" />\n  <img src="https://circleci.com/gh/jlugao/sauron-rule-engine/tree/master.svg?style=svg" />\n  <a href=\'https://coveralls.io/github/jlugao/sauron-rule-engine?branch=master\'><img src=\'https://coveralls.io/repos/github/jlugao/sauron-rule-engine/badge.svg?branch=master&service=github\' alt=\'Coverage Status\' /></a>\n\n<img alt="GitHub" src="https://img.shields.io/github/license/jlugao/sauron-rule-engine.svg?style=plastic">\n  <a href="https://twitter.com/joaovoce">\n    <img alt="Twitter: joaovoce" src="https://img.shields.io/twitter/follow/joaovoce.svg?style=social" target="_blank" />\n  </a>\n</p>\n\n[![Build Status](https://dev.azure.com/jlugao/Sauron%20Engine/_apis/build/status/jlugao.sauron-rule-engine?branchName=master)](https://dev.azure.com/jlugao/Sauron%20Engine/_build/latest?definitionId=1&branchName=master)\n\n> A simple rule engine to be used in python, it is based on simple rules and actions that can be chained with each other. The idea is to run the rule processor on events and have it mutate data or trigger actions\n\nHeavily inspired on FastAPI. We use type annotations in our engine so that we can export data to other systems or frontends to convey what conditions and actions are possible using that engine\n\n## Install\n\n```sh\npip install sauron-rule-engine\n```\n\n## Concepts\n\nSauron rule engine is based on custom functions that can be called by a rule.\n\n### Condition\n\nCondition to be satisfied in order for the actions to run, they can take some or no parameters at all\nMultiple conditions can be chained in order to create more complex ones, currently all chained conditions must be satisfied\n\n### Action\n\nAn Action is the intented result. Usually they are there to mutate state or trigger/schedule other kinds of actions in your system. Actions can also be chained and will run in order.\n\n### Rule\n\nA Rule is a dict or json string containing the conditions and actions and the arguments they should be run with. Usually those rules will be built by a frontend to match complex and adaptable business rules from your customer\n\n## Use it\n\nA simple example of the usage\n\n```python\nfrom sauron.rule_engine import RuleEngine\n\nengine = RuleEngine()\n\n\n@engine.condition("First Condition")\ndef first_condition(session,lower_number: int = 10, greater_number: int = 20) -> bool:\n    """\n    Checks if first number is lower than the first\n    - lower_number: Number expected to be low\n    - higher_number: Number expected to be high\n    """\n    return lower_number < greater_number\n\n\n@engine.condition()\ndef second_condition(session):\n    """\n    Takes no argument and always returns True\n    """\n    return True\n\n\n@engine.action("The Action")\ndef print_the_equation(\n    session, lower_number: int = 10, greater_number: int = 20\n) -> None:\n    """\n    Prints a statement Asserting that the first number is lower than the second number\n    - lower_number: Number expected to be low\n    - higher_number: Number expected to be high\n    """\n    print(f"{lower_number} < {greater_number}")\n\n\nrule = {\n    "conditions": [\n        {\n            "name": "first_condition",\n            "args": {"lower_number": 3, "greater_number": 10},\n        }\n    ],\n    "actions": [\n        {\n            "name": "print_the_equation",\n            "args": {"lower_number": 3, "greater_number": 10},\n        }\n    ],\n}\n\n\nengine.run(rule)\n```\n\n## Choices Fields\n\nChoices fields are supported through python\'s built-in Enum type. Example:\n\n```python\nfrom sauron.rule_engine import RuleEngine\nfrom enum import Enum\n\nclass Color(str, Enum):\n    red = "R"\n    green = "G"\n    blue = "B"\n\n\n@engine.condition("is it red?")\ndef is_red(session, color: Color) -> bool:\n    """\n    Checks if the color is red\n    """\n    return color == color.red\n\n```\n\n## Export Conditions and Actions\n\nYou can use the function export_metadata to export your data in a dict or as a json string (just pass `json=True`). Here is an Example and the output:\n\n```python\nfrom sauron_rule_engine.rule_engine import RuleEngine\nfrom enum import Enum\n\nengine = RuleEngine()\n\n\n@engine.condition("First Condition")\ndef first_condition(lower_number: int = 10, greater_number: int = 20) -> bool:\n    """\n    Checks if first number is lower than the first\n    - lower_number: Number expected to be low\n    - higher_number: Number expected to be high\n    """\n    return lower_number < greater_number\n\n\n@engine.condition()\ndef second_condition():\n    """\n    Takes no argument and always returns True\n    """\n    return True\n\n\n@engine.action("The Action")\ndef print_the_equation(\n    lower_number: int = 10, greater_number: int = 20\n) -> None:\n    """\n    Prints a statement Asserting that the first number is lower than the second number\n    - lower_number: Number expected to be low\n    - higher_number: Number expected to be high\n    """\n    print(f"{lower_number} < {greater_number}")\n\n\nclass Color(str, Enum):\n    red = "R"\n    green = "G"\n    blue = "B"\n\n\n@engine.condition("is it red?")\ndef is_red(color: Color) -> bool:\n    """\n    Checks if the color is red\n    """\n    return color == color.red\n\n\nmetadata = engine.export_metadata(json=True)\nprint(metadata)\n\n```\n\nResults in the following json to be served to your frontend:\n\n```json\n{\n  "actions": {\n    "print_the_equation": {\n      "args": {\n        "lower_number": { "default": 10, "type": "int", "choices": null },\n        "greater_number": { "default": 20, "type": "int", "choices": null }\n      },\n      "doc": "Prints a statement Asserting that the first number is lower than the second number\\n- lower_number: Number expected to be low\\n- higher_number: Number expected to be high",\n      "name": "The Action"\n    }\n  },\n  "conditions": {\n    "first_condition": {\n      "args": {\n        "lower_number": { "default": 10, "type": "int", "choices": null },\n        "greater_number": { "default": 20, "type": "int", "choices": null }\n      },\n      "doc": "Checks if first number is lower than the first\\n- lower_number: Number expected to be low\\n- higher_number: Number expected to be high",\n      "name": "First Condition"\n    },\n    "second_condition": {\n      "args": {},\n      "doc": "Takes no argument and always returns True",\n      "name": "second_condition"\n    },\n    "is_red": {\n      "args": {\n        "color": {\n          "default": null,\n          "type": "Color",\n          "choices": ["red", "green", "blue"]\n        }\n      },\n      "doc": "Checks if the color is red",\n      "name": "is it red?"\n    }\n  }\n}\n```\n\n## Sessions\n\nResults are stored in a result stack inside the session, so that jobs can share data with each other. \n\n\n## More Features coming to town\n\n- Support pydantic types\n- Support for complex types with hints to the frontend (like a range for an int type)\n\n## Contribute\n\n- We need all the help we can get. Please read [CONTRIBUTE.md](CONTRIBUTING.md) for instructions\n\n## Author\n\n👤 **João Ricardo Lhullier Lugão**\n\n- Twitter: [@joaovoce](https://twitter.com/joaovoce)\n- Github: [@jlugao](https://github.com/jlugao)\n\n## Show your support\n\nGive a ⭐️ if this project helped you!\n\n---\n\n_This README was generated with ❤️ by [readme-md-generator](https://github.com/kefranabg/readme-md-generator)_\n',
    'author': 'jlugao',
    'author_email': 'joaolhullier@gmail.com',
    'url': 'https://github.com/jlugao/sauron-rule-engine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
